//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%

#ifndef CREATEBINNEDPLOTGROUPS_H
#define CREATEBINNEDPLOTGROUPS_H 1
#include <XSPlot/Plot/PlotTypes.h>
#include <XSPlot/Plot/PlotSettings.h>
#include <list>

// PlotGroupCreator
#include <XSPlot/Plot/PlotGroupCreator.h>


// Enumerate the different type of binned plots that can be calculated.
// Note that the RATE type works for plots such as residuals as well as data
typedef enum {RATE, COUNTS, UNFOLDED, FITSTAT, DELCHI} BinnedPlotType;


class CreateBinnedPlotGroups : public PlotGroupCreator  //## Inherits: <unnamed>%4AB7CB9102EF
{

  public:
      CreateBinnedPlotGroups (BinnedPlotType type);
      virtual ~CreateBinnedPlotGroups();

  virtual std::vector<PlotGroup*> createPlotGroups (const PlotSettings& settings,
						    const IntegerVector& plotGroupNumbers);
      BinnedPlotType binnedPlotType () const;

    // Additional Public Declarations

  protected:
    // Additional Protected Declarations

  private:
      CreateBinnedPlotGroups(const CreateBinnedPlotGroups &right);
      CreateBinnedPlotGroups & operator=(const CreateBinnedPlotGroups &right);

      void itemizePlotGroupModels ();
      PlotGroup* initializePlotGroup (size_t groupIndex, std::list<SpectralData*>& groupSpectra);
      bool getNextPlotBin (PlotGroup* gr, int npts, std::list<SpectralData*>& groupSpectra, int& latestChannel);
      void integrateModelFluxes (std::list<SpectralData*>& spectra, Real eMin, Real eMax, PlotGroup* group, int point);
      void setError (Real& error, Real& backError, PlotSettings::PlotErrMode errorType, Real areaSum,
		     Real backAreaSum, Real count, Real backgroundCount, Real bscaleRatio,
		     bool allUsingChiSquare);
      static void initEffectiveAreas (std::list<SpectralData*>& spectraInGroup);
      static void clearEffectiveAreas (std::list<SpectralData*>& spectraInGroup);

    // Additional Private Declarations

  private: //## implementation
    // Data Members for Class Attributes
      const BinnedPlotType m_type;
      ModelPositionInfo m_modelPositionDirectory;
      SpectraModelHolder m_modelsForSpectra;
      const PlotSettings* m_settings;

    // Additional Implementation Declarations

};

// Class CreateBinnedPlotGroups 

inline BinnedPlotType CreateBinnedPlotGroups::binnedPlotType () const
{
  return m_type;
}


#endif
